# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import importlib.resources
import os

import PIL.Image

from torchvision.transforms.functional import center_crop, to_tensor

from ptbench.data.datamodule import CachingDataModule
from ptbench.data.image_utils import remove_black_borders
from ptbench.data.split import JSONDatabaseSplit
from ptbench.data.typing import DatabaseSplit
from ptbench.data.typing import RawDataLoader as _BaseRawDataLoader
from ptbench.data.typing import Sample
from ptbench.utils.rc import load_rc

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(_BaseRawDataLoader):
    """A specialized raw-data-loader for the Shenzen dataset."""

    datadir: str
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = load_rc().get(
            CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir)
        )

    def sample(self, sample: tuple[str, int]) -> Sample:
        """Loads a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.


        Returns
        -------
            The sample representation
        """
        # images from TBPOC are encoded as grayscale JPEGs, no need to
        # call convert("L") here.
        image = PIL.Image.open(os.path.join(self.datadir, sample[0]))
        image = remove_black_borders(image)
        tensor = to_tensor(image)
        tensor = center_crop(tensor, min(*tensor.shape[1:]))

        # use the code below to view generated images
        # from torchvision.transforms.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return tensor, dict(label=sample[1], name=sample[0])  # type: ignore[arg-type]

    def label(self, sample: tuple[str, int]) -> int:
        """Loads a single image sample label from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.


        Returns
        -------
            The integer label associated with the sample
        """
        return sample[1]


def make_split(basename: str) -> DatabaseSplit:
    """Returns a database split for the TB-POC database."""

    return JSONDatabaseSplit(
        importlib.resources.files(__name__.rsplit(".", 1)[0]).joinpath(basename)
    )


class DataModule(CachingDataModule):
    """TB-POC dataset for computer-aided diagnosis.

    * Database reference: [TB-POC-2018]_
    * Original resolution (height x width or width x height): 2048 x 2500 pixels
      or 2500 x 2048 pixels

    Data specifications:

    * Raw data input (on disk):

      * JPEG 8-bit Grayscale images
      * resolution: fixed to one of the cases above

    * Output image:

      * Transforms:

        * Load raw grayscale jpeg with :py:mod:`PIL`
        * Remove black borders
        * Convert to torch tensor
        * Torch center cropping to get square image

      * Final specifications:

        * Grayscale, encoded as a single plane tensor, 32-bit floats,
          square with varying resolutions (2048 x 2048 being the maximum),
          but also depending on black borders' sizes on the input image.
        * Labels: 0 (healthy), 1 (active tuberculosis)
    """

    def __init__(self, split_filename: str):
        super().__init__(
            database_split=make_split(split_filename),
            raw_data_loader=RawDataLoader(),
        )
