# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Defines functionality for the evaluation of predictions."""

import contextlib
import itertools
import logging
import typing

from collections.abc import Iterable, Iterator

import matplotlib.figure
import numpy
import numpy.typing
import sklearn.metrics
import tabulate

from matplotlib import pyplot as plt

from ..models.typing import BinaryPrediction

logger = logging.getLogger(__name__)


def eer_threshold(predictions: Iterable[BinaryPrediction]) -> float:
    """Calculates the (approximate) threshold leading to the equal error rate.

    Parameters
    ----------
    predictions
        An iterable of multiple
        :py:data:`.models.typing.BinaryPrediction`'s.


    Returns
    -------
        The EER threshold value.
    """
    from scipy.interpolate import interp1d
    from scipy.optimize import brentq

    y_scores = [k[2] for k in predictions]
    y_labels = [k[1] for k in predictions]

    fpr, tpr, thresholds = sklearn.metrics.roc_curve(y_labels, y_scores)

    eer = brentq(lambda x: 1.0 - x - interp1d(fpr, tpr)(x), 0.0, 1.0)
    return float(interp1d(fpr, thresholds)(eer))


def _get_centered_maxf1(
    f1_scores: numpy.typing.NDArray, thresholds: numpy.typing.NDArray
):
    """Return the centered max F1 score threshold when multiple thresholds give
    the same max F1 score.

    Parameters
    ----------
    f1_scores
        1D array of f1 scores
    thresholds
        1D array of thresholds

    Returns
    -------
        A tuple with the maximum F1-score and the "centered" threshold.
    """
    maxf1 = f1_scores.max()
    maxf1_indices = numpy.where(f1_scores == maxf1)[0]

    # If multiple thresholds give the same max F1 score
    if len(maxf1_indices) > 1:
        mean_maxf1_index = int(round(numpy.mean(maxf1_indices)))
    else:
        mean_maxf1_index = maxf1_indices[0]

    return maxf1, thresholds[mean_maxf1_index]


def maxf1_threshold(predictions: Iterable[BinaryPrediction]) -> float:
    """Calculates the threshold leading to the maximum F1-score on a precision-
    recall curve.

    Parameters
    ----------
    predictions
        An iterable of multiple
        :py:data:`.models.typing.BinaryPrediction`'s.


    Returns
    -------
        The threshold value leading to the maximum F1-score on the provided set
        of predictions.
    """
    y_scores = [k[2] for k in predictions]
    y_labels = [k[1] for k in predictions]

    precision, recall, thresholds = sklearn.metrics.precision_recall_curve(
        y_labels, y_scores
    )

    numerator = 2 * recall * precision
    denom = recall + precision
    f1_scores = numpy.divide(
        numerator, denom, out=numpy.zeros_like(denom), where=(denom != 0)
    )

    _, maxf1_threshold = _get_centered_maxf1(f1_scores, thresholds)
    return maxf1_threshold


def _score_plot(
    labels: numpy.typing.NDArray,
    scores: numpy.typing.NDArray,
    title: str,
    threshold: float,
) -> matplotlib.figure.Figure:
    """Plots the normalized score distributions for all systems.

    Parameters
    ----------
    labels
        True labels (negatives and positives) for each entry in ``scores``
    scores
        Likelihoods provided by the classification model, for each sample
    title
        Title of the plot.
    threshold
        Shows where the threshold is in the figure


    Returns
    -------
        A single (matplotlib) plot containing the score distribution, ready to
        be saved to disk or displayed.
    """

    fig, ax = plt.subplots(1, 1)
    assert isinstance(fig, matplotlib.figure.Figure)
    ax = typing.cast(plt.Axes, ax)  # gets editor to behave

    # Here, we configure the "style" of our plot
    ax.set_xlim([0, 1])
    ax.set_title(title)
    ax.set_xlabel("Score")
    ax.set_ylabel("Normalized count")
    ax.grid(linestyle="--", linewidth=1, color="gray", alpha=0.2)

    # Only show ticks on the left and bottom spines
    ax.spines.right.set_visible(False)
    ax.spines.top.set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    positives = scores[labels > 0.5]
    negatives = scores[labels < 0.5]

    ax.hist(positives, bins="auto", label="positives", density=True, alpha=0.7)
    ax.hist(negatives, bins="auto", label="negatives", density=True, alpha=0.7)

    # Adds threshold line (dotted red)
    ax.axvline(
        threshold,  # type: ignore
        color="red",
        lw=2,
        alpha=0.75,
        ls="dotted",
        label="threshold",
    )

    # Adds a nice legend
    ax.legend(
        title="Max F1-scores",
        fancybox=True,
        framealpha=0.7,
    )

    # Makes sure the figure occupies most of the possible space
    fig.tight_layout()

    return fig


def run_binary(
    name: str,
    predictions: Iterable[BinaryPrediction],
    threshold_a_priori: float | None = None,
) -> tuple[
    dict[str, typing.Any],
    dict[str, matplotlib.figure.Figure],
    dict[str, typing.Any],
]:
    """Runs inference and calculates measures for binary classification.

    Parameters
    ----------
    name
        The name of subset to load.
    predictions
        A list of predictions to consider for measurement
    threshold_a_priori
        A threshold to use, evaluated *a priori*, if must report single values.
        If this value is not provided, a *a posteriori* threshold is calculated
        on the input scores.  This is a biased estimator.


    Returns
    -------
        A tuple containing the following entries:

        * summary: A dictionary containing the performance summary on the
          specified threshold
        * figures: A dictionary of generated standalone figures
        * curves: A dictionary containing curves that can potentially be combined
          with other prediction lists to make aggregate plots.
    """

    y_scores = numpy.array([k[2] for k in predictions])  # likelihoods
    y_labels = numpy.array([k[1] for k in predictions])  # integers
    neg_label = y_labels.min()
    pos_label = y_labels.max()

    use_threshold = threshold_a_priori
    if use_threshold is None:
        use_threshold = maxf1_threshold(predictions)
        logger.warning(
            f"User did not pass an *a priori* threshold for the evaluation "
            f"of split `{name}`.  Using threshold a posteriori (biased) with value "
            f"`{use_threshold:.4f}`"
        )

    y_predictions = numpy.where(y_scores >= use_threshold, pos_label, neg_label)

    # point measures on threshold
    summary = dict(
        split=name,
        threshold=use_threshold,
        threshold_a_posteriori=(threshold_a_priori is None),
        precision=sklearn.metrics.precision_score(
            y_labels, y_predictions, pos_label=pos_label
        ),
        recall=sklearn.metrics.recall_score(
            y_labels, y_predictions, pos_label=pos_label
        ),
        specificity=sklearn.metrics.recall_score(
            y_labels, y_predictions, pos_label=neg_label
        ),
        accuracy=sklearn.metrics.accuracy_score(y_labels, y_predictions),
        f1_score=sklearn.metrics.f1_score(
            y_labels, y_predictions, pos_label=pos_label
        ),
    )

    # figures: score distributions
    figures = dict(
        scores=_score_plot(
            y_labels,
            y_scores,
            f"Score distribution (split: {name})",
            use_threshold,
        ),
    )

    # curves: ROC and precision recall
    curves = dict(
        roc=sklearn.metrics.roc_curve(y_labels, y_scores, pos_label=pos_label),
        precision_recall=sklearn.metrics.precision_recall_curve(
            y_labels, y_scores, pos_label=pos_label
        ),
    )

    return summary, figures, curves


def aggregate_summaries(
    data: typing.Sequence[typing.Mapping[str, typing.Any]], fmt: str
) -> str:
    """Tabulates summaries from multiple splits.

    This function can properly tabulate the various summaries produced for all
    the splits in a prediction database.


    Parameters
    ----------
    data
        An iterable over all summary data collected
    fmt
        One of the formats supported by `python-tabulate
        <https://pypi.org/project/tabulate/>`_.


    Returns
    -------
        A string containing the tabulated information
    """

    headers = list(data[0].keys())
    table = [[k[h] for h in headers] for k in data]
    return tabulate.tabulate(table, headers, tablefmt=fmt, floatfmt=".3f")


def aggregate_roc(
    data: typing.Mapping[str, typing.Any],
    title: str = "ROC",
) -> matplotlib.figure.Figure:
    """Aggregates ROC curves from multiple splits.

    This function produces a single ROC plot for multiple curves generated per
    split.


    Parameters
    ----------
    data
        A dictionary mapping split names to ROC curve data produced by
        :py:func:sklearn.metrics.roc_curve`.

    title
        The title of the plot.

    Returns
    -------
        A figure, containing the aggregated ROC plot.
    """
    fig, ax = plt.subplots(1, 1)
    assert isinstance(fig, matplotlib.figure.Figure)

    # Names and bounds
    ax.set_xlabel("1 - specificity")
    ax.set_ylabel("Sensitivity")
    ax.set_xlim([0.0, 1.0])
    ax.set_ylim([0.0, 1.0])
    ax.set_title(title)

    # we should see some of ax 1 ax
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["left"].set_position(("data", -0.015))
    ax.spines["bottom"].set_position(("data", -0.015))

    ax.grid(linestyle="--", linewidth=1, color="gray", alpha=0.2)

    plt.tight_layout()

    lines = ["-", "--", "-.", ":"]
    colors = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
    ]
    colorcycler = itertools.cycle(colors)
    linecycler = itertools.cycle(lines)

    legend = []

    for name, (fpr, tpr, _) in data.items():
        # plots roc curve
        _auc = sklearn.metrics.auc(fpr, tpr)
        label = f"{name} (AUC={_auc:.2f})"
        color = next(colorcycler)
        style = next(linecycler)

        (line,) = ax.plot(fpr, tpr, color=color, linestyle=style)
        legend.append((line, label))

    if len(legend) > 1:
        ax.legend(
            [k[0] for k in legend],
            [k[1] for k in legend],
            loc="lower right",
            fancybox=True,
            framealpha=0.7,
        )

    return fig


@contextlib.contextmanager
def _precision_recall_canvas() -> (
    Iterator[tuple[matplotlib.figure.Figure, matplotlib.figure.Axes]]
):
    """Generates a canvas to draw precision-recall curves.

    Works like a context manager, yielding a figure and an axes set in which
    the precision-recall curves should be added to.  The figure already
    contains F1-ISO lines and is preset to a 0-1 square region.  Once the
    context is finished, ``fig.tight_layout()`` is called.


    Yields
    ------
    figure
        The figure that should be finally returned to the user
    axes
        An axis set where to precision-recall plots should be added to
    """

    fig, axes1 = plt.subplots(1)
    assert isinstance(fig, matplotlib.figure.Figure)
    assert isinstance(axes1, matplotlib.figure.Axes)

    # Names and bounds
    axes1.set_xlabel("Recall")
    axes1.set_ylabel("Precision")
    axes1.set_xlim([0.0, 1.0])
    axes1.set_ylim([0.0, 1.0])

    axes1.grid(linestyle="--", linewidth=1, color="gray", alpha=0.2)
    axes2 = axes1.twinx()

    # Annotates plot with F1-score iso-lines
    f_scores = numpy.linspace(0.1, 0.9, num=9)
    tick_locs = []
    tick_labels = []
    for f_score in f_scores:
        x = numpy.linspace(0.01, 1)
        y = f_score * x / (2 * x - f_score)
        plt.plot(x[y >= 0], y[y >= 0], color="green", alpha=0.1)
        tick_locs.append(y[-1])
        tick_labels.append("%.1f" % f_score)
    axes2.tick_params(axis="y", which="both", pad=0, right=False, left=False)
    axes2.set_ylabel("iso-F", color="green", alpha=0.3)
    axes2.set_ylim([0.0, 1.0])
    axes2.yaxis.set_label_coords(1.015, 0.97)
    axes2.set_yticks(tick_locs)  # notice these are invisible
    for k in axes2.set_yticklabels(tick_labels):
        k.set_color("green")
        k.set_alpha(0.3)
        k.set_size(8)

    # we should see some of axes 1 axes
    axes1.spines["right"].set_visible(False)
    axes1.spines["top"].set_visible(False)
    axes1.spines["left"].set_position(("data", -0.015))
    axes1.spines["bottom"].set_position(("data", -0.015))

    # we shouldn't see any of axes 2 axes
    axes2.spines["right"].set_visible(False)
    axes2.spines["top"].set_visible(False)
    axes2.spines["left"].set_visible(False)
    axes2.spines["bottom"].set_visible(False)

    # yield execution, lets user draw precision-recall plots, and the legend
    # before tighteneing the layout
    yield fig, axes1

    plt.tight_layout()


def aggregate_pr(
    data: typing.Mapping[str, typing.Any],
    title: str = "Precision-Recall Curve",
) -> matplotlib.figure.Figure:
    """Aggregates PR curves from multiple splits.

    This function produces a single Precision-Recall plot for multiple curves
    generated per split. The plot will be annotated with F1-score iso-lines (in
    which the F1-score maintains the same value).


    Parameters
    ----------
    data
        A dictionary mapping split names to ROC curve data produced by
        :py:func:sklearn.metrics.precision_recall_curve`.
    title
        The title of the plot.


    Returns
    -------
        A figure, containing the aggregated PR plot.
    """

    lines = ["-", "--", "-.", ":"]
    colors = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
    ]
    colorcycler = itertools.cycle(colors)
    linecycler = itertools.cycle(lines)

    with _precision_recall_canvas() as (fig, axes):
        axes.set_title(title)
        legend = []

        for name, (prec, recall, _) in data.items():
            _auc = sklearn.metrics.auc(recall, prec)
            label = f"{name} (AUC={_auc:.2f})"
            color = next(colorcycler)
            style = next(linecycler)

            (line,) = axes.plot(recall, prec, color=color, linestyle=style)
            legend.append((line, label))

        if len(legend) > 1:
            axes.legend(
                [k[0] for k in legend],
                [k[1] for k in legend],
                loc="lower left",
                fancybox=True,
                framealpha=0.7,
            )

    return fig
