# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os
import pathlib
import typing

import lightning.pytorch
import matplotlib.pyplot
import numpy
import numpy.typing
import PIL.Image
import PIL.ImageColor
import PIL.ImageDraw
import torchvision.transforms.functional

from tqdm import tqdm

from ...config.data.tbx11k.datamodule import BoundingBox, BoundingBoxes

logger = logging.getLogger(__name__)


def _overlay_saliency_map(
    image: PIL.Image.Image,
    saliencies: numpy.typing.NDArray[numpy.double],
    colormap: typing.Literal[  # we accept any "Sequential" colormap from mpl
        "viridis",
        "plasma",
        "inferno",
        "magma",
        "cividis",
        "Greys",
        "Purples",
        "Blues",
        "Greens",
        "Oranges",
        "Reds",
        "YlOrBr",
        "YlOrRd",
        "OrRd",
        "PuRd",
        "RdPu",
        "BuPu",
        "GnBu",
        "PuBu",
        "YlGnBu",
        "PuBuGn",
        "BuGn",
        "YlGn",
    ],
    image_weight: float,
) -> PIL.Image.Image:
    """Creates an overlayed represention of the saliency map on the original
    image.

    This is a slightly modified version of the show_cam_on_image implementation in:
    https://github.com/jacobgil/pytorch-grad-cam, but uses matplotlib instead
    of opencv.


    Parameters
    ----------
    image
        The input imge that will be overlayed with the saliency map
    saliencies
        The saliency map that will be overlaid on the (raw) image
    colormap
        The name of the (matplotlib) colormap to be used
    image_weight
        The final result is ``image_weight * image + (1-image_weight) *
        saliency_map``.


    Returns
    -------
        A modified version of the input ``image`` with the overlaid saliency
        map.
    """

    image_array = numpy.array(image, dtype=numpy.float32) / 255.0

    assert image_array.shape[:2] == saliencies.shape, (
        f"The shape of the saliency map ({saliencies.shape}) is different "
        f"from the shape of the input image ({image_array.shape[:2]})."
    )

    assert (
        saliencies.max() <= 1
    ), f"The input saliency map should be in the range [0, 1] (max={saliencies.max()})"

    assert (
        image_weight > 0 and image_weight < 1
    ), f"image_weight should be in the range [0, 1], but got {image_weight}"

    heatmap = matplotlib.pyplot.cm.get_cmap(colormap)(saliencies)

    # For pixels where the mask is zero, the original image pixels are being
    # used without a mask.
    result = numpy.where(
        saliencies[..., numpy.newaxis] == 0,
        image_array,
        (image_weight * image_array) + ((1 - image_weight) * heatmap[:, :, :3]),
    )

    return PIL.Image.fromarray((result * 255).astype(numpy.uint8), "RGB")


def _overlay_bounding_box(
    image: PIL.Image.Image,
    bbox: BoundingBox,
    color: str,
    width: int,
) -> PIL.Image.Image:
    """Draws ground-truth on the input image.

    Parameters
    ----------
    image
        The input imge that will be overlayed with the saliency map
    bbox
        The bounding box to draw on the input image
    color
        The color to use for drawing the bounding box. Any of the colours in
        :any:`PIL.ImageColor.colormap` are accepted.
    width
        The width of the bounding box, in pixels.  A larger value creates a
        bounding box that is thicker, towards the outside of the boxed area.


    Returns
    -------
        A modified version of the input ``image`` with the ground-truth drawn
        on the top.
    """

    draw = PIL.ImageDraw.Draw(image)
    draw.rectangle(
        (bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax),
        outline=PIL.ImageColor.getrgb(color),
        width=width,
    )
    return image


def _process_sample(
    raw_data: numpy.typing.NDArray[numpy.double],
    saliencies: numpy.typing.NDArray[numpy.double],
    ground_truth: BoundingBoxes,
) -> PIL.Image.Image:
    """Generates an overlayed representation of the original sample and
    saliency maps.

    Parameters
    ----------
    raw_data
        The raw data representing the input sample that will be overlayed with
        saliency maps and annotations
    saliencies
        The saliency map recovered from the model, that will be inprinted on
        the raw_data
    ground_truth
        Ground-truth annotations that may be inprinted on the final image


    Returns
    -------
        An image with the original raw data overlayed with the different
        elements as selected by the user.
    """

    # we need a colour image to eventually overlay a (coloured) saliency map on
    # the top, draw rectangles and other annotations in coulour.  So, we force
    # it right up front.
    retval = torchvision.transforms.functional.to_pil_image(raw_data).convert(
        "RGB"
    )

    retval = _overlay_saliency_map(
        retval, saliencies, colormap="plasma", image_weight=0.5
    )

    for k in ground_truth:
        retval = _overlay_bounding_box(retval, k, color="green", width=2)

    return retval


def run(
    datamodule: lightning.pytorch.LightningDataModule,
    input_folder: pathlib.Path,
    target_label: int,
    output_folder: pathlib.Path,
    show_groundtruth: bool,
    threshold: float,
):
    """Overlays saliency maps on CXR to output final images with heatmaps.

    Parameters
    ----------
    datamodule
        The lightning datamodule to iterate on.
    input_folder
        Directory in which the saliency maps are stored for a specific
        visualization type.
    target_label
        The label to target for evaluating interpretability metrics. Samples
        contining any other label are ignored.
    output_folder
        Directory in which the resulting visualisations will be saved.
    show_groundtruth
        If set, inprint ground truth labels over the original image and
        saliency maps.
    threshold : float
        The pixel values above ``threshold``% of max value are kept in the
        original saliency map.  Everything else is set to zero.  The value
        proposed on [SCORECAM-2020]_ is 0.2.  Use this value if unsure.
    """

    for dataset_name, dataset_loader in datamodule.predict_dataloader().items():
        logger.info(
            f"Generating visualisations for samples at dataset `{dataset_name}`..."
        )

        for sample in tqdm(
            dataset_loader, desc="batches", leave=False, disable=None
        ):
            name = str(sample[1]["name"][0])
            label = int(sample[1]["label"].item())
            data = sample[0][0]

            if label != target_label:
                # no visualisation was generated
                continue

            saliencies = numpy.load(
                input_folder / pathlib.Path(name).with_suffix(".npy")
            )
            saliencies[saliencies < (threshold * saliencies.max())] = 0

            # TODO: This is very specific to the TBX11k system for labelling
            # regions of interest.  We need to abstract from this to support more
            # datasets and other ways to annotate.
            if show_groundtruth:
                ground_truth = sample[1].get("bounding_boxes", BoundingBoxes())
            else:
                ground_truth = BoundingBoxes()

            # we fully process this entry
            image = _process_sample(
                data,
                saliencies,
                ground_truth,
            )

            # Save image
            output_file_path = output_folder / pathlib.Path(name).with_suffix(
                ".png"
            )
            os.makedirs(output_file_path.parent, exist_ok=True)
            image.save(output_file_path)
