# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import typing

import lightning.pytorch as pl
import torch
import torch.nn
import torch.optim.optimizer
import torch.utils.data
import torchvision.models as models
import torchvision.transforms

from ..data.typing import TransformSequence
from .separate import separate
from .transforms import RGB
from .typing import Checkpoint

logger = logging.getLogger(__name__)


class Densenet(pl.LightningModule):
    """Densenet-121 module.

    Parameters
    ----------
    train_loss
        The loss to be used during the training.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.

    validation_loss
        The loss to be used for validation (may be different from the training
        loss).  If extra-validation sets are provided, the same loss will be
        used throughout.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.

    optimizer_type
        The type of optimizer to use for training

    optimizer_arguments
        Arguments to the optimizer after ``params``.

    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.

    pretrained
        If set to True, loads pretrained model weights during initialization,
        else trains a new model.

    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        train_loss: torch.nn.Module = torch.nn.BCEWithLogitsLoss(),
        validation_loss: torch.nn.Module | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] = {},
        augmentation_transforms: TransformSequence = [],
        pretrained: bool = False,
        num_classes: int = 1,
    ):
        super().__init__()

        self.name = "densenet-121"
        self.num_classes = num_classes

        # image is probably large, resize first to get memory usage down
        self.model_transforms = [
            torchvision.transforms.Resize(512, antialias=True),
            RGB(),
        ]

        self._train_loss = train_loss
        self._validation_loss = (
            validation_loss if validation_loss is not None else train_loss
        )
        self._optimizer_type = optimizer_type
        self._optimizer_arguments = optimizer_arguments

        self._augmentation_transforms = torchvision.transforms.Compose(
            augmentation_transforms
        )

        self.pretrained = pretrained

        # Load pretrained model
        if not pretrained:
            weights = None
        else:
            logger.info(f"Loading pretrained {self.name} model weights")
            weights = models.DenseNet121_Weights.DEFAULT

        self.model_ft = models.densenet121(weights=weights)

        # Adapt output features
        self.model_ft.classifier = torch.nn.Linear(
            self.model_ft.classifier.in_features, self.num_classes
        )

    def forward(self, x):
        x = self.normalizer(x)  # type: ignore

        x = self.model_ft(x)

        return x

    def on_save_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Called by Lightning to restore your model.

        If you saved something with on_save_checkpoint() this is your chance to
        restore this.

        Parameters
        ----------
        checkpoint
            Loaded checkpoint
        """
        checkpoint["normalizer"] = self.normalizer

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Called by Lightning when saving a checkpoint to give you a chance to
        store anything else you might want to save.

        Parameters
        ----------
        checkpoint
            Loaded checkpoint
        """
        logger.info("Restoring normalizer from checkpoint.")
        self.normalizer = checkpoint["normalizer"]

    def set_normalizer(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Initializes the normalizer for the current model.

        This function is NOOP if ``pretrained = True`` (normalizer set to
        imagenet weights, during contruction).

        Parameters
        ----------
        dataloader
            A torch Dataloader from which to compute the mean and std.
            Will not be used if the model is pretrained.
        """
        if self.pretrained:
            from .normalizer import make_imagenet_normalizer

            logger.warning(
                f"ImageNet pre-trained {self.name} model - NOT "
                f"computing z-norm factors from train dataloader. "
                f"Using preset factors from torchvision."
            )
            self.normalizer = make_imagenet_normalizer()
        else:
            from .normalizer import make_z_normalizer

            logger.info(
                f"Uninitialised {self.name} model - "
                f"computing z-norm factors from train dataloader."
            )
            self.normalizer = make_z_normalizer(dataloader)

    def training_step(self, batch, _):
        images = batch[0]
        labels = batch[1]["label"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # Forward pass on the network
        outputs = self(self._augmentation_transforms(images))

        return self._train_loss(outputs, labels.float())

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        images = batch[0]
        labels = batch[1]["label"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # data forwarding on the existing network
        outputs = self(images)

        return self._validation_loss(outputs, labels.float())

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        outputs = self(batch[0])
        probabilities = torch.sigmoid(outputs)
        return separate((probabilities, batch[1]))

    def configure_optimizers(self):
        return self._optimizer_type(
            self.parameters(), **self._optimizer_arguments
        )
