# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import typing

import lightning.pytorch as pl
import torch
import torch.nn
import torch.nn.functional as F
import torch.optim.optimizer
import torch.utils.data
import torchvision.transforms

from ..data.typing import TransformSequence
from .separate import separate
from .transforms import Grayscale
from .typing import Checkpoint

logger = logging.getLogger(__name__)


class Pasa(pl.LightningModule):
    """Implementation of CNN by Pasa.

    Simple CNN for classification based on paper by [PASA-2019]_.

    This network has a linear output.  You should use losses with ``WithLogit``
    instead of cross-entropy versions when training.


    Parameters
    ----------
    train_loss
        The loss to be used during the training.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.

    validation_loss
        The loss to be used for validation (may be different from the training
        loss).  If extra-validation sets are provided, the same loss will be
        used throughout.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.

    optimizer_type
        The type of optimizer to use for training

    optimizer_arguments
        Arguments to the optimizer after ``params``.

    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.

    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        train_loss: torch.nn.Module = torch.nn.BCEWithLogitsLoss(),
        validation_loss: torch.nn.Module | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] = {},
        augmentation_transforms: TransformSequence = [],
        num_classes: int = 1,
    ):
        super().__init__()

        self.name = "pasa"
        self.num_classes = num_classes

        # image is probably large, resize first to get memory usage down
        self.model_transforms = [
            torchvision.transforms.Resize(512, antialias=True),
            Grayscale(),
        ]

        self._train_loss = train_loss
        self._validation_loss = (
            validation_loss if validation_loss is not None else train_loss
        )
        self._optimizer_type = optimizer_type
        self._optimizer_arguments = optimizer_arguments

        self._augmentation_transforms = torchvision.transforms.Compose(
            augmentation_transforms
        )

        # First convolution block
        self.fc1 = torch.nn.Conv2d(1, 4, (3, 3), (2, 2), (1, 1))
        self.fc2 = torch.nn.Conv2d(4, 16, (3, 3), (2, 2), (1, 1))
        self.fc3 = torch.nn.Conv2d(1, 16, (1, 1), (4, 4))

        self.batchNorm2d_4 = torch.nn.BatchNorm2d(4)
        self.batchNorm2d_16 = torch.nn.BatchNorm2d(16)
        self.batchNorm2d_16_2 = torch.nn.BatchNorm2d(16)

        # Second convolution block
        self.fc4 = torch.nn.Conv2d(16, 24, (3, 3), (1, 1), (1, 1))
        self.fc5 = torch.nn.Conv2d(24, 32, (3, 3), (1, 1), (1, 1))
        self.fc6 = torch.nn.Conv2d(
            16, 32, (1, 1), (1, 1)
        )  # Original stride (2, 2)

        self.batchNorm2d_24 = torch.nn.BatchNorm2d(24)
        self.batchNorm2d_32 = torch.nn.BatchNorm2d(32)
        self.batchNorm2d_32_2 = torch.nn.BatchNorm2d(32)

        # Third convolution block
        self.fc7 = torch.nn.Conv2d(32, 40, (3, 3), (1, 1), (1, 1))
        self.fc8 = torch.nn.Conv2d(40, 48, (3, 3), (1, 1), (1, 1))
        self.fc9 = torch.nn.Conv2d(
            32, 48, (1, 1), (1, 1)
        )  # Original stride (2, 2)

        self.batchNorm2d_40 = torch.nn.BatchNorm2d(40)
        self.batchNorm2d_48 = torch.nn.BatchNorm2d(48)
        self.batchNorm2d_48_2 = torch.nn.BatchNorm2d(48)

        # Fourth convolution block
        self.fc10 = torch.nn.Conv2d(48, 56, (3, 3), (1, 1), (1, 1))
        self.fc11 = torch.nn.Conv2d(56, 64, (3, 3), (1, 1), (1, 1))
        self.fc12 = torch.nn.Conv2d(
            48, 64, (1, 1), (1, 1)
        )  # Original stride (2, 2)

        self.batchNorm2d_56 = torch.nn.BatchNorm2d(56)
        self.batchNorm2d_64 = torch.nn.BatchNorm2d(64)
        self.batchNorm2d_64_2 = torch.nn.BatchNorm2d(64)

        # Fifth convolution block
        self.fc13 = torch.nn.Conv2d(64, 72, (3, 3), (1, 1), (1, 1))
        self.fc14 = torch.nn.Conv2d(72, 80, (3, 3), (1, 1), (1, 1))
        self.fc15 = torch.nn.Conv2d(
            64, 80, (1, 1), (1, 1)
        )  # Original stride (2, 2)

        self.batchNorm2d_72 = torch.nn.BatchNorm2d(72)
        self.batchNorm2d_80 = torch.nn.BatchNorm2d(80)
        self.batchNorm2d_80_2 = torch.nn.BatchNorm2d(80)

        self.pool2d = torch.nn.MaxPool2d(
            (3, 3), (2, 2)
        )  # Pool after conv. block
        self.dense = torch.nn.Linear(
            80, self.num_classes
        )  # Fully connected layer

    def forward(self, x):
        x = self.normalizer(x)  # type: ignore

        # First convolution block
        _x = x
        x = F.relu(self.batchNorm2d_4(self.fc1(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_16(self.fc2(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_16_2(self.fc3(_x)))) / 2  # Parallel
        x = self.pool2d(x)  # Pooling

        # Second convolution block
        _x = x
        x = F.relu(self.batchNorm2d_24(self.fc4(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_32(self.fc5(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_32_2(self.fc6(_x)))) / 2  # Parallel
        x = self.pool2d(x)  # Pooling

        # Third convolution block
        _x = x
        x = F.relu(self.batchNorm2d_40(self.fc7(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_48(self.fc8(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_48_2(self.fc9(_x)))) / 2  # Parallel
        x = self.pool2d(x)  # Pooling

        # Fourth convolution block
        _x = x
        x = F.relu(self.batchNorm2d_56(self.fc10(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_64(self.fc11(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_64_2(self.fc12(_x)))) / 2  # Parallel
        x = self.pool2d(x)  # Pooling

        # Fifth convolution block
        _x = x
        x = F.relu(self.batchNorm2d_72(self.fc13(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_80(self.fc14(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_80_2(self.fc15(_x)))) / 2  # Parallel
        # no pooling

        # Global average pooling
        x = torch.mean(x.view(x.size(0), x.size(1), -1), dim=2)

        # Dense layer
        x = self.dense(x)

        # x = F.log_softmax(x, dim=1) # 0 is batch size

        return x

    def on_save_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Called by Lightning to restore your model.

        If you saved something with on_save_checkpoint() this is your chance to
        restore this.

        Parameters
        ----------
        checkpoint
            Loaded checkpoint
        """
        checkpoint["normalizer"] = self.normalizer

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Called by Lightning when saving a checkpoint to give you a chance to
        store anything else you might want to save.

        Parameters
        ----------
        checkpoint
            Loaded checkpoint
        """
        logger.info("Restoring normalizer from checkpoint.")
        self.normalizer = checkpoint["normalizer"]

    def set_normalizer(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Initializes the input normalizer for the current model.

        Parameters
        ----------
        dataloader
            A torch Dataloader from which to compute the mean and std
        """
        from .normalizer import make_z_normalizer

        logger.info(
            f"Uninitialised {self.name} model - "
            f"computing z-norm factors from train dataloader."
        )
        self.normalizer = make_z_normalizer(dataloader)

    def training_step(self, batch, _):
        images = batch[0]
        labels = batch[1]["label"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # Forward pass on the network
        outputs = self(self._augmentation_transforms(images))

        return self._train_loss(outputs, labels.float())

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        images = batch[0]
        labels = batch[1]["label"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # data forwarding on the existing network
        outputs = self(images)

        return self._validation_loss(outputs, labels.float())

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        outputs = self(batch[0])
        probabilities = torch.sigmoid(outputs)
        return separate((probabilities, batch[1]))

    def configure_optimizers(self):
        return self._optimizer_type(
            self.parameters(), **self._optimizer_arguments
        )
