# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click

from clapper.click import AliasedGroup, verbosity_option
from clapper.logging import setup

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def _get_raw_databases() -> dict[str, dict[str, str]]:
    """Returns a list of all supported (raw) databases.

    Returns
    -------
    d
        Dictionary where keys are database names, and values are dictionaries
        containing two string keys:

        * ``module``: the full Pythonic module name (e.g.
        ``ptbench.data.montgomery``)
        * ``datadir``: points to the user-configured data directory for the
        current dataset, if set, or ``None`` otherwise.
    """
    import importlib
    import pkgutil

    from ..config import data
    from ..utils.rc import load_rc

    user_configuration = load_rc()

    retval = {}
    for k in pkgutil.iter_modules(data.__path__):
        for j in pkgutil.iter_modules(
            [next(iter(data.__path__)) + f"/{k.name}"]
        ):
            if j.name == "datamodule":
                # this is a submodule that can read raw data files
                module = importlib.import_module(
                    f".{j.name}", data.__package__ + f".{k.name}"
                )
                if hasattr(module, "CONFIGURATION_KEY_DATADIR"):
                    retval[k.name] = dict(
                        module=module.__name__.rsplit(".", 1)[0],
                        datadir=user_configuration.get(
                            module.CONFIGURATION_KEY_DATADIR
                        ),
                    )
                else:
                    retval[k.name] = dict(module=module.__name__)

    return retval


@click.group(cls=AliasedGroup)
def database() -> None:
    """Commands for listing and verifying databases installed."""
    pass


@database.command(
    epilog="""Examples:

\b
    1. To install a database, set up its data directory ("datadir").  For
       example, to setup access to Montgomery files you downloaded locally at
       the directory "/path/to/montgomery/files", edit the RC file (typically
       ``$HOME/.config/ptbench.toml``), and add a line like the following:

       .. code:: toml

          [datadir]
          montgomery = "/path/to/montgomery/files"

       .. note::

          This setting **is** case-sensitive.

\b
    2. List all raw databases supported (and configured):

       .. code:: sh

          $ ptbench database list

""",
)
@verbosity_option(logger=logger, expose_value=False)
def list():
    """Lists all supported and configured databases."""
    config = _get_raw_databases()

    click.echo("Available databases:")
    for k, v in config.items():
        if "datadir" not in v:
            # this database does not have a "datadir"
            continue

        if v["datadir"] is not None:
            click.secho(f'- {k} ({v["module"]}): "{v["datadir"]}"', fg="green")
        else:
            click.echo(f'- {k} ({v["module"]}): NOT installed')


@database.command(
    epilog="""Examples:

    1. Check if all files from the split 'montgomery-f0' of the Montgomery
       database can be loaded:

       .. code:: sh

          ptbench datamodule check -vv montgomery-f0

""",
)
@click.argument(
    "split",
    nargs=1,
)
@click.option(
    "--limit",
    "-l",
    help="Limit check to the first N samples in each split dataset, making the "
    "check sensibly faster.  Set it to zero (default) to check everything.",
    required=True,
    type=click.IntRange(0),
    default=0,
)
@verbosity_option(logger=logger, expose_value=False)
def check(split, limit):
    """Checks file access on one or more datamodules."""
    import importlib.metadata
    import sys

    click.secho(f"Checking split `{split}`...", fg="yellow")
    try:
        module = importlib.metadata.entry_points(group="ptbench.config")[
            split
        ].module
    except KeyError:
        raise Exception(f"Could not find database split `{split}`")

    datamodule = importlib.import_module(module).datamodule

    datamodule.model_transforms = []  # should be done before setup()
    datamodule.batch_size = 1  # ensure one sample is loaded at a time
    datamodule.setup("predict")  # sets up all datasets

    loaders = datamodule.predict_dataloader()

    errors = 0
    for k, loader in loaders.items():
        if limit == 0:
            click.secho(
                f"Checking all samples of dataset `{k}` at split `{split}`...",
                fg="yellow",
            )
            loader_limit = sys.maxsize
        else:
            click.secho(
                f"Checking first {limit} samples of dataset "
                f"`{k}` at split `{split}`...",
                fg="yellow",
            )
            loader_limit = limit
        # the for loop will trigger raw data loading (ie. user code), protect
        # it
        try:
            for i, batch in enumerate(loader):
                if loader_limit == 0:
                    break
                logger.info(
                    f"{batch[1]['name'][0]}: "
                    f"{[s for s in batch[0][0].shape]}@{batch[0][0].dtype}"
                )
                loader_limit -= 1
        except Exception:
            logger.exception(f"Unable to load batch {i} in dataset {k}")
            errors += 1

    if not errors:
        click.secho(
            f"OK! No errors were reported for database split `{split}`.",
            fg="green",
        )
    else:
        click.secho(
            f"Found {errors} errors loading datamodule `{split}`.", fg="red"
        )
