# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click

from clapper.click import ConfigCommand, ResourceOption, verbosity_option
from clapper.logging import setup

from .train import reusable_options as training_options

# avoids X11/graphical desktop requirement when creating plots
__import__("matplotlib").use("agg")

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


@click.command(
    entry_point_group="ptbench.config",
    cls=ConfigCommand,
    epilog="""Examples:

\b
  1. Trains a pasa model with montgomery dataset, on the CPU, for only two
     epochs, then runs inference and evaluation on stock datasets, report
     performance as a table and a figure:

     .. code:: sh

        $ ptbench experiment -vv pasa montgomery --epochs=2
""",
)
@training_options
@verbosity_option(logger=logger, cls=ResourceOption)
@click.pass_context
def experiment(
    ctx,
    model,
    output_folder,
    epochs,
    batch_size,
    batch_chunk_count,
    drop_incomplete_batch,
    datamodule,
    validation_period,
    device,
    cache_samples,
    seed,
    parallel,
    monitoring_interval,
    resume_from,
    balance_classes,
    **_,
):
    """Runs a complete experiment, from training, to prediction and evaluation.

    This script is just a wrapper around the individual scripts for training,
    running prediction, evaluating and comparing model performance.  It
    organises the output in a preset way::

        \b
       └─ <output-folder>/
          ├── command.sh
          ├── model/  # the generated model will be here
          ├── predictions.json  # the prediction outputs for the sets
          └── evaluation/  # the outputs of the evaluations for the sets
    """
    from .utils import save_sh_command

    save_sh_command(output_folder / "command.sh")

    # training
    logger.info("Started training")

    from .train import train

    train_output_folder = output_folder / "model"
    ctx.invoke(
        train,
        model=model,
        output_folder=train_output_folder,
        epochs=epochs,
        batch_size=batch_size,
        batch_chunk_count=batch_chunk_count,
        drop_incomplete_batch=drop_incomplete_batch,
        datamodule=datamodule,
        validation_period=validation_period,
        device=device,
        cache_samples=cache_samples,
        seed=seed,
        parallel=parallel,
        monitoring_interval=monitoring_interval,
        resume_from=resume_from,
        balance_classes=balance_classes,
    )

    logger.info("Ended training")

    logger.info("Started train analysis")
    from .train_analysis import train_analysis

    logdir = train_output_folder / "logs"
    output_pdf = train_output_folder / "trainlog.pdf"
    ctx.invoke(
        train_analysis,
        logdir=logdir,
        output=output_pdf,
    )

    logger.info("Ended train analysis")

    logger.info("Started predicting")

    from .predict import predict

    predictions_output = output_folder / "predictions.json"

    ctx.invoke(
        predict,
        output=predictions_output,
        model=model,
        datamodule=datamodule,
        device=device,
        weight=train_output_folder,
        batch_size=batch_size,
        parallel=parallel,
    )

    logger.info("Ended predicting")

    logger.info("Started evaluating")

    from .evaluate import evaluate

    evaluations_folder = output_folder / "evaluation"

    ctx.invoke(
        evaluate,
        output_folder=evaluations_folder,
        predictions=predictions_output,
        threshold="validation",
    )

    logger.info("Ended evaluating")
