# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import pathlib

import click

from clapper.click import ConfigCommand, ResourceOption, verbosity_option
from clapper.logging import setup

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


@click.command(
    entry_point_group="ptbench.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Generates visualizations in form of heatmaps from existing saliency maps for a dataset configuration:

   .. code:: sh

      ptbench saliency view -vv pasa tbx11k-v1-healthy-vs-atb --input-folder=parent_folder/gradcam/ --output-folder=path/to/visualizations
""",
)
@click.option(
    "--model",
    "-m",
    help="A lightining module instance implementing the network to be used for applying the necessary data transformations.",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--datamodule",
    "-d",
    help="A lighting data module containing the training, validation and test sets.",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--input-folder",
    "-i",
    help="Path to the folder containing the saliency maps for a specific visualization type.",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="visualizations",
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="Path where to store the ROAD scores (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="visualizations",
    cls=ResourceOption,
)
@click.option(
    "--show-groundtruth/--no-show-groundtruth",
    "-G/-g",
    help="""If set, visualizations for ground truth labels will be generated.
    Only works for datasets with bounding boxes.""",
    is_flag=True,
    default=False,
    cls=ResourceOption,
)
@click.option(
    "--threshold",
    "-t",
    help="""The pixel values above ``threshold`` % of max value are kept in the
    original saliency map.  Everything else is set to zero.  The value proposed
    on [SCORECAM-2020]_ is 0.2.  Use this value if unsure.""",
    show_default=True,
    required=True,
    default=0.2,
    type=click.FloatRange(min=0, max=1),
    cls=ResourceOption,
)
@verbosity_option(logger=logger, cls=ResourceOption, expose_value=False)
def view(
    model,
    datamodule,
    input_folder,
    output_folder,
    show_groundtruth,
    threshold,
    **_,
) -> None:
    """Generates heatmaps for input CXRs based on existing saliency maps."""

    from ...engine.saliency.viewer import run
    from ..utils import save_sh_command

    assert (
        input_folder != output_folder
    ), "Output folder must not be the same as the input folder."

    assert not str(output_folder).startswith(
        str(input_folder)
    ), "Output folder must not be a subdirectory of the input folder."

    logger.info(f"Output folder: {output_folder}")
    os.makedirs(output_folder, exist_ok=True)

    save_sh_command(output_folder / "command.sh")

    datamodule.set_chunk_size(1, 1)
    datamodule.drop_incomplete_batch = False
    # datamodule.cache_samples = cache_samples
    # datamodule.parallel = parallel
    datamodule.model_transforms = model.model_transforms

    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    run(
        datamodule=datamodule,
        input_folder=input_folder,
        target_label=1,
        output_folder=output_folder,
        show_groundtruth=show_groundtruth,
        threshold=threshold,
    )
