# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import click

from clapper.click import verbosity_option
from clapper.logging import setup

# avoids X11/graphical desktop requirement when creating plots
__import__("matplotlib").use("agg")

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def create_figures(
    data: dict[str, tuple[list[int], list[float]]],
    groups: list[str] = [
        "total-execution-time-seconds",
        "loss/*",
        "learning-rate",
        "memory-used-GB/cpu/*" "rss-GB/cpu/*",
        "vms-GB/cpu/*",
        "num-open-files/cpu/*",
        "num-processes/cpu/*",
        "percent-usage/cpu/*",
        # nvidia gpu
        "memory-percent/gpu/*",
        "memory-used-GB/gpu/*",
        "memory-free-GB/gpu/*",
        "memory-free-GB/gpu/*",
        "percent-usage/gpu/*",
    ],
) -> list:
    """Generates figures for each metric in the dataframe.

    Each row of the dataframe correspond to an epoch and each column to a metric.
    It is assumed that some metric names are of the form <metric>/<subset>.
    All subsets for a metric will be displayed on the same figure.


    Parameters
    ----------

    data:
        A dictionary where keys represent all scalar names, and values
        correspond to a tuple that contains an array with epoch numbers (when
        values were taken), when the monitored values themselves.  These lists
        are pre-sorted by epoch number.
    groups:
        A list of scalar globs we are interested on the existing tensorboard
        data, for plotting.  Values with multiple matches are drawn on the same
        plot.  Values that do not exist are ignored.


    Returns
    -------

    figures:
        List of matplotlib figures, one per metric.
    """
    import fnmatch
    import typing

    import matplotlib.pyplot as plt

    from matplotlib.ticker import MaxNLocator

    figures = []

    for group in groups:
        curves = {k: data[k] for k in fnmatch.filter(data.keys(), group)}

        if len(curves) == 0:
            continue

        fig, ax = plt.subplots(1, 1)
        ax = typing.cast(plt.Axes, ax)
        fig = typing.cast(plt.Figure, fig)

        if len(curves) == 1:
            # there is only one curve, just plot it
            title, (epochs, values) = next(iter(curves.items()))
            ax.plot(epochs, values)

        else:
            # this is an aggregate plot, name things consistently
            labels = {k: k[len(group) - 1 :] for k in curves.keys()}
            title = group.rstrip("*").rstrip("/")
            for key, (epochs, values) in curves.items():
                ax.plot(epochs, values, label=labels[key])
            ax.legend(loc="best")

        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.set_title(title)
        ax.set_xlabel("Epoch")
        ax.set_ylabel(title)

        ax.grid(alpha=0.3)
        fig.tight_layout()
        figures.append(fig)

    return figures


@click.command(
    epilog="""Examples:

\b
    1. Analyzes a training log and produces various plots:

       .. code:: sh

          ptbench train-analysis -vv results/logs
""",
)
@click.argument(
    "logdir",
    type=click.Path(dir_okay=True, exists=True, path_type=pathlib.Path),
)
@click.option(
    "--output",
    "-o",
    help="Name of the output file to dump (multi-page PDF)",
    required=True,
    show_default=True,
    default="trainlog.pdf",
    type=click.Path(dir_okay=False, file_okay=True, path_type=pathlib.Path),
)
@verbosity_option(logger=logger, expose_value=False)
def train_analysis(
    logdir: pathlib.Path,
    output: pathlib.Path,
) -> None:
    """Creates a plot for each metric in the training logs and saves them in a
    pdf file."""

    import matplotlib.pyplot as plt

    from matplotlib.backends.backend_pdf import PdfPages

    from ..utils.tensorboard import scalars_to_dict

    data = scalars_to_dict(logdir)

    output.parent.mkdir(parents=True, exist_ok=True)

    with PdfPages(output) as pdf:
        for figure in create_figures(data):
            pdf.savefig(figure)
            plt.close(figure)
