# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import importlib.metadata
import logging
import os
import pathlib
import shutil
import sys
import time

logger = logging.getLogger(__name__)


def save_sh_command(path: pathlib.Path) -> None:
    """Records command-line to reproduce this script.

    This function can record the current command-line used to call the script
    being run.  It creates an executable ``bash`` script setting up the current
    working directory and activating a conda environment, if needed.  It
    records further information on the date and time the script was run and the
    version of the package.


    Parameters
    ----------

    path
        Path to a file where the commands to reproduce the current run will be
        recorded.  Parent directories will be created if they do not exist. An
        existing copy will be backed-up if it exists.
    """

    logger.info(f"Writing command-line for reproduction at `{path}`...")

    # create parent directories
    path.parent.mkdir(parents=True, exist_ok=True)

    # backup if exists
    if path.exists():
        backup = path.parent / (path.name + "~")
        shutil.copy(path, backup)

    # write the file
    package = __name__.split(".", 1)[0]
    version = importlib.metadata.version(package)

    with path.open("w") as f:
        f.write("#!/usr/bin/env sh\n")
        f.write(f"# date: {time.asctime()}\n")
        f.write(f"# version: {version} ({package})\n")
        f.write(f"# platform: {sys.platform}\n")
        f.write("\n")
        args = []
        for k in sys.argv:
            if " " in k:
                args.append(f'"{k}"')
            else:
                args.append(k)
        if os.environ.get("CONDA_DEFAULT_ENV") is not None:
            f.write(f"# conda activate {os.environ['CONDA_DEFAULT_ENV']}\n")
        f.write(f"# cd {os.path.realpath(os.curdir)}\n")
        f.write(" ".join(args) + "\n")

    # make it executable
    path.chmod(0o755)
